"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ErrorRegistry = void 0;
const BaseRegistryError_1 = require("./error-types/BaseRegistryError");
const BaseError_1 = require("./error-types/BaseError");
/**
 * Contains the definitions for High and Low Level Errors and
 * generates custom errors from those definitions.
 */
class ErrorRegistry {
    constructor(highLvErrors, lowLvErrors, config = {}) {
        this.highLevelErrors = highLvErrors;
        this.lowLevelErrors = {};
        this.classNameHighLevelNameMap = {};
        this.highLevelErrorClasses = {};
        this._config = config;
        this._newErrorContext = null;
        Object.keys(highLvErrors).forEach(name => {
            this.classNameHighLevelNameMap[highLvErrors[name].className] = name;
        });
        // populate the lowLevelErrors dictionary
        Object.keys(lowLvErrors).forEach(type => {
            const errCode = lowLvErrors[type];
            errCode.type = type;
            this.lowLevelErrors[type] = errCode;
        });
    }
    /**
     * Gets the definition of a High Level Error
     * @param {string} highLvErrName
     */
    getHighLevelError(highLvErrName) {
        return this.highLevelErrors[highLvErrName];
    }
    /**
     * Gets the definition of a Low Level Error
     * @param {string} lowLvErrName
     */
    getLowLevelError(lowLvErrName) {
        return this.lowLevelErrors[lowLvErrName];
    }
    /**
     * Gets the class definition of a High Level Error
     * @param highLvErrName
     */
    getClass(highLvErrName) {
        const highLevelDef = this.getHighLevelError(highLvErrName);
        if (!highLevelDef) {
            throw new Error(`High level error not defined: ${highLvErrName}`);
        }
        // Create class definition of the highLevelError and cache
        if (!this.highLevelErrorClasses[highLvErrName]) {
            // https://stackoverflow.com/questions/33605775/es6-dynamic-class-names
            const C = class extends BaseRegistryError_1.BaseRegistryError {
            };
            Object.defineProperty(C, 'name', { value: highLevelDef.className });
            this.highLevelErrorClasses[highLvErrName] = C;
        }
        return this.highLevelErrorClasses[highLvErrName];
    }
    /**
     * Compares an instance of an object to a specified High Level Error
     */
    instanceOf(a, highLvErrName) {
        return a instanceof this.getClass(highLvErrName);
    }
    /**
     * Creates an instance of a High Level Error, without a Low Level Error
     * attached to it.
     * @param {string} highLvErrName
     * @param {string} [message] Error message. If not defined and the high level error has the
     * message property defined, will use that instead. If the high level error message is not defined,
     * then will use the high level error id instead for the message.
     */
    newBareError(highLvErrName, message) {
        const hlErrDef = this.highLevelErrors[highLvErrName];
        if (!message && hlErrDef.message) {
            message = hlErrDef.message;
        }
        else if (!message) {
            message = hlErrDef.code.toString();
        }
        const C = this.getClass(highLvErrName);
        const err = new C(this.getHighLevelError(highLvErrName), {
            message
        }, this._config.baseErrorConfig);
        if (typeof this._config.onCreateError === 'function') {
            this._config.onCreateError(err);
        }
        this.reformatTrace(err);
        return this.assignErrContext(err);
    }
    /**
     * Creates an instance of a High Level Error, with a Low Level Error
     * attached to it.
     * @param {string} highLvErrName
     * @param {string} lowLvErrName
     */
    newError(highLvErrName, lowLvErrName) {
        if (!this.lowLevelErrors[lowLvErrName]) {
            throw new Error(`Low level error not defined: ${lowLvErrName}`);
        }
        const C = this.getClass(highLvErrName);
        const err = new C(this.getHighLevelError(highLvErrName), this.getLowLevelError(lowLvErrName), this._config.baseErrorConfig);
        if (typeof this._config.onCreateError === 'function') {
            this._config.onCreateError(err);
        }
        this.reformatTrace(err);
        return this.assignErrContext(err);
    }
    /**
     * Updates the stack trace to remove the error registry entry:
     * "at ErrorRegistry.newError" and related entries
     */
    reformatTrace(err) {
        const stack = err.stack.split('\n');
        stack.splice(1, 1);
        err.stack = stack.join('\n');
    }
    /**
     * If a new error context is defined, then set the values
     */
    assignErrContext(err) {
        if (!this._newErrorContext) {
            return err;
        }
        if (this._newErrorContext.metadata) {
            err.withMetadata(this._newErrorContext.metadata);
        }
        if (this._newErrorContext.safeMetadata) {
            err.withSafeMetadata(this._newErrorContext.safeMetadata);
        }
        return err;
    }
    /**
     * Deserializes data into an error
     * @param {string} data JSON.parse()'d error object from
     * BaseError#toJSON() or BaseError#toJSONSafe()
     * @param {DeserializeOpts} [opts] Deserialization options
     */
    fromJSON(data, opts) {
        if (typeof data !== 'object') {
            throw new Error(`fromJSON(): Data is not an object.`);
        }
        // data.name is the class name - we need to resolve it to the name of the high level class definition
        const errorName = this.classNameHighLevelNameMap[data.name];
        // use the lookup results to see if we can get the class definition of the high level error
        const highLevelDef = this.getHighLevelError(errorName);
        let err = null;
        // Can deserialize into an custom error instance class
        if (highLevelDef) {
            // get the class for the error type
            const C = this.getClass(errorName);
            err = C.fromJSON(data, opts);
        }
        else {
            err = BaseError_1.BaseError.fromJSON(data, opts);
        }
        return err;
    }
    /**
     * Creates a new error registry instance that will include / set specific data
     * for each new error created.
     *
     * Memory overhead should be trivial as the internal *references* of the existing
     * error registry instance properties is copied over to the new instance.
     */
    withContext(context) {
        const registry = new ErrorRegistry({}, {});
        Object.keys(this).forEach(property => {
            // copy over references vs creating completely new entries
            registry[property] = this[property];
        });
        registry._newErrorContext = context;
        return registry;
    }
}
exports.ErrorRegistry = ErrorRegistry;
//# sourceMappingURL=ErrorRegistry.js.map